-- nscrm3d.lua

local setmetatable = setmetatable
local assert = assert
local type = type

local NSSleep = NSSleep

local function alpha2as(alpha)
  -- NScripter 0-255 to ActionScript 0-100
  return alpha * 100.0 / 255.0
end

local _G = _G

module("nscrm3d")


Render = {
  Every = -1,
  None = 0,
  Force = 1,
}

function new(nsswf)
  local genCommand = function(args)
    return nsswf:buildCallFunction("doCommand", args)
  end
  local doCommand = function(args)
    return genCommand(args)()
  end
  return setmetatable({nsswf = nsswf, interval = nsswf:getInterval(), doCommand = doCommand, genCommand = genCommand}, {__index = _M})
end


function lm3d(self, id, url, x, y, z, alpha)
  assert(type(id) == "number")
  assert(type(url) == "string")
  if y then y = -y end -- x,y,z : left->right, top->bottom, front->back
  local success = self.doCommand{command = "model.load", id = id, url = url, x = x, y = y, z = z, alpha = alpha and alpha2as(alpha) or nil}
  if not success then return false end
  local loaded = self.genCommand{command = "model.loaded", id = id}
  while not loaded() do
    NSSleep(self.interval)
  end
  return true
end


function cm3d(self, id)
  assert(type(id) == "number")
  return self.doCommand{command = "model.clear", id = id}
end


function mm3d(self, id, dx, dy, dz, dalpha)
  assert(type(id) == "number")
  if dy then dy = -dy end -- x,y,z : left->right, top->bottom, front->back
  return self.doCommand{command = "model.moveBy", id = id, x = dx, y = dy, z = dz, alpha = dalpha and alpha2as(dalpha) or nil}
end


function amm3d(self, id, x, y, z, alpha)
  assert(type(id) == "number")
  if y then y = -y end -- x,y,z : left->right, top->bottom, front->back
  return self.doCommand{command = "model.moveTo", id = id, x = x, y = y, z = z, alpha = alpha and alpha2as(alpha) or nil}
end


function vm3d(self, id, visible)
  assert(type(id) == "number")
  return self.doCommand{command = "model.visible", id = id, visible = visible}
end


function rm3d(self, id, dx, dy, dz)
  assert(type(id) == "number")
  if dy then dy = -dy end -- x,y,z : clockwise in looking to plusward.
  return self.doCommand{command = "model.rotateBy", id = id, x = dx, y = dy, z = dz}
end


function arm3d(self, id, x, y, z)
  assert(type(id) == "number")
  if y then y = -y end -- x,y,z : clockwise in looking to plusward.
  return self.doCommand{command = "model.rotateTo", id = id, x = x, y = y, z = z}
end


function m3dmotionload(self, id, url, append)
  assert(type(id) == "number")
  assert(type(url) == "string")
  local success = self.doCommand{command = "motion.load", id = id, url = url, append = append}
  if not success then return false end
  local loaded = self.genCommand{command = "motion.loaded", id = id}
  while not loaded() do
    NSSleep(self.interval)
  end
  return true
end


function m3dmotionclear(self, id)
  assert(type(id) == "number")
  return self.doCommand{command = "motion.clear", id = id}
end


function m3dmotionplay(self, id)
  assert(type(id) == "number")
  return self.doCommand{command = "motion.play", id = id}
end


function m3dmotionstop(self, id)
  assert(type(id) == "number")
  return self.doCommand{command = "motion.stop", id = id}
end


function m3dcameraset(self, vars)
  assert(type(vars) == "table")
  return self.doCommand{command = "camera.setVars", vars = vars}
end


function m3dcameraadd(self, vars)
  assert(type(vars) == "table")
  return self.doCommand{command = "camera.addVars", vars = vars}
end


function m3dcamerareset(self)
  return self.doCommand{command = "camera.reset"}
end


function m3drender(self, value)
  value = value or Render.Force
  self.doCommand{command = "renderOn", value = value}
end


function m3dtrace(self, value)
  if type(value) == "number" then value = value ~= 0 end
  self.doCommand{command = "trace", value = value}
end
