-- swf.lua
-- @author senzogawa
-- @release For nsswf.dll 1.2.0.2

local function layermessage(num, command, ...)
  local args = {...}
  local params = #args == 0 and "" or table.concat(args, ",")
  NSSetStrValue(0, string.format('%s(%s)', command, params))
  NSExec(string.format('_layermessage %d,$0', num))
--  NSExec(string.format('_layermessage %d,"%s(%s)"', num, command, params))
end


local function getret_int()
  local tnum = NSGetIntValue(0)
  NSExec('_getret %0')
  local num = NSGetIntValue(0)
  NSSetIntValue(0, tnum)
  return num
end


local function getret_str()
  local tstr = NSGetStrValue(0)
  NSExec('_getret $0')
  local str = NSGetStrValue(0)
  NSSetStrValue(0, tstr)
  return str
end

local table = require 'table'
local string = require 'string'
local io = require 'io'

local setmetatable = setmetatable
local assert = assert
local tostring = tostring
local tonumber = tonumber
local ipairs = ipairs
local pairs = pairs
local type = type
local unpack = unpack
local next = next

local NSExec = NSExec
local NSSpLoad = NSSpLoad
local NSSpVisible = NSSpVisible
local NSSpClear = NSSpClear

--- nsswfɂSWFĐx܂B
module("swf")

DllPath = "nsswf.dll"

DefaultSprite = 10
DefaultLayer = 10
DefaultInterval = 20

QualityLow = 0
QualityHigh = 1

Success = 1
Failure = 0

Playing = 1
Stopped = 0

Enabled = 1
Disabled = 0


--- ftHg̃XvCgԍAC[ԍAC^[ow肵܂B
-- @param sprite XvCgԍ
-- @param layer C[ԍ
-- @param interval C^[o(~b)
function setDefault(sprite, layer, interval)
  assert(sprite and type(sprite) == "number")
  assert(layer and type(layer) == "number")
  assert(interval and type(interval) == "number")
  DefaultSprite = sprite
  DefaultLayer = layer
  DefaultInterval = interval
end


--- nsswf܂B
-- nil̏ꍇ̓ftHglp܂B
-- @param layer C[ԍ
-- @param interval C^[o
-- @param dll DLL̃pX
function init(layer, interval, dll)
  layer = layer or DefaultLayer
  interval = interval or DefaultInterval
  dll = dll or DllPath
  NSExec(string.format('_setlayer %d,%d,"%s"', layer, interval, dll))
end


--- nsswfp̃CX^X𐶐܂B
-- nil̏ꍇ̓ftHglp܂B
-- @param sprite XvCgԍ
-- @param layer C[ԍ
-- @param visible \(ftHg͔\)
function create(sprite, layer, visible)
  sprite = sprite or DefaultSprite
  layer = layer or DefaultLayer
  visible = not not visible
  assert(sprite)
  assert(layer)
  NSSpLoad(sprite, "*"..layer)
  NSSpVisible(sprite, visible)
  return setmetatable({sprite = sprite, layer = layer}, {__index = _M})
end


-- methods

function open(self, file, async)
  assert(file)
  if async then
    layermessage(self.layer, "SwfBeginOpen", file)
    self.async = true
  else
    layermessage(self.layer, "SwfOpen", file)
    self.async = false
    self.opened = true
  end
  return getret_int()
end


function sync(self)
  if not self.async then return end
  layermessage(self.layer, "SwfEndOpen")
  self.opened = true
  self.async = false
end


function sendMessage(self, command, ...)
  self:sync()
  layermessage(self.layer, command, ...)
end


function play(self)
  self:sendMessage("SwfPlay")
  return getret_int()
end


function stop(self)
  self:sendMessage("SwfStop")
  return getret_int()
end


function pause(self)
  self:sendMessage("SwfPause")
  return getret_int()
end


function close(self)
  if not self.opened then return end
  self:sendMessage("SwfClose")
  self.opened = false
end


function dispose(self)
  self:close()
  NSSpClear(self.sprite)
end


function move(self, x, y)
  assert(x or y)
  if not y then
    self:sendMessage("SwfSetPosX", x)
  elseif not x then
    self:sendMessage("SwfSetPosY", y)
  else
    self:sendMessage("SwfMove", x, y)
  end
  return getret_int()
end


function getPos(self)
  self:sendMessage("SwfGetPosX")
  local x = getret_int()
  self:sendMessage("SwfGetPosY")
  local y = getret_int()
  return x, y
end


function resize(self, width, height)
  assert(width or height)
  if not height then
    self:sendMessage("SwfSetWidth", width)
  elseif not width then
    self:sendMessage("SwfSetHeight", height)
  else
    self:sendMessage("SwfResize", width, height)
  end
  return getret_int()
end


function getSize(self)
  self:sendMessage("SwfGetWidth")
  local width = getret_int()
  self:sendMessage("SwfGetHeight")
  local height = getret_int()
  return width, height
end


function show(self)
  self:sendMessage("SwfShow")
  NSSpVisible(self.sprite, true)
end


function hide(self)
  self:sendMessage("SwfHide")
  NSSpVisible(self.sprite, false)
end


function setVisible(self, visible)
  local on = visible and "ON" or "OFF"
  self:sendMessage("SwfSetVisible", on)
  NSSpVisible(self.sprite, visible)
  return getret_int()
end


function getVisible(self)
  self:sendMessage("SwfGetVisible")
  return getret_int()
end


function isVisible(self)
  return self:getVisible() == Enabled
end


function setLoop(self, loop)
  local on = loop and "ON" or "OFF"
  self:sendMessage("SwfSetLoop", on)
  return getret_int()
end


function getLoop(self)
  self:sendMessage("SwfGetLoop")
  return getret_int()
end


function isLoop(self)
  return self:getLoop() == Enabled
end


function setQuality(self, quality)
  assert(quality == QualityLow or quality == QualityHigh)
  self:sendMessage("SwfSetQuality", quality)
  return getret_int()
end


function getQuality(self)
  self:sendMessage("SwfGetQuality")
  return getret_int()
end


function setCurrentFrame(self, frame)
  assert(frame)
  self:sendMessage("SwfSetCurrentFrame", frame)
  return getret_int()
end


function getCurrentFrame(self)
  self:sendMessage("SwfGetCurrentFrame")
  return getret_int()
end


function getTotalFrame(self)
  self:sendMessage("SwfGetTotalFrames")
  return getret_int()
end


function isPlaying(self)
  self:sendMessage("SwfIsPlaying")
  return getret_int() == Playing
end


function setVariable(self, name, value)
  self:sendMessage("SwfSetVariable", name, tostring(value))
  return getret_int()
end


function setFlashVars(self, vars)
  local list = {}
  for name,value in pairs(vars) do
    table.insert(list, name .. "=" .. tostring(value))
  end
  self:sendMessage("SwfSetFlashVars", table.concat(list, "&"))
  return getret_int()
end


function getVariable(self, name)
  self:sendMessage("SwfGetVariable", name)
  if getret_int() == 0 then return nil end
  return getret_str()
end


function getVariableAsNumber(self, name)
  local var = self:getVariable(name)
  if var == nil then return nil end
  return tonumber(var)
end


function getVariableAsBoolean(self, name)
  local var = self:getVariable(name)
  if not var then return false end
  return var ~= "false"
end


function getOptions(self, var_names)
  local options = {}
  options.x, options.y = self:getPos()
  options.width, options.height = self:getSize()
  options.visible = self:isVisible()
  options.loop = self:isLoop()
  options.quality = self:getQuality()
  options.frame = self:getCurrentFrame()
  if not var_names then return options end
  local vars = {}
  for _,name in ipairs(var_names) do
    vars[name] = self:getVariable(name)
  end
  options.vars = vars
end


local function ensureValueAsType(type, value)
  if type == "string" then
    return value
  elseif type == "number" then
    return tonumber(value)
  elseif type == "null" then
    return nil
  elseif type == "true" then
    return true
  elseif type == "false" then
    return false
  end
  return nil
end


local function parseReturnFunction(str)
  local returns = {}
  local pos = 1
  while true do
    local type, close = str:match("<(%w+)(/?)>", pos)
    if type == nil then break end
    if #close ~= 0 then
      -- empty
      table.insert(returns, ensureValueAsType(type))
      pos = pos + #type + #close + 2
    else
      local type, value = str:match("<(%w+)>(.*)</%1>", pos)
      table.insert(returns, ensureValueAsType(type, value))
      pos = pos + #type * 2 + #value + 5
    end
  end
  return unpack(returns)
end


function callFunction(self, str)
  self:sendMessage("SwfCallFunction", str)
  if getret_int() == 0 then return nil end
  return parseReturnFunction(getret_str())
end


local function buildCallFunctionArg(arg)
  -- http://www.adobe.com/livedocs/flex/3_jp/html/19_External_Interface_08.html#549525
  local arg_type = type(arg)
  if arg_type == "number" then
    return "<number>" .. arg .. "</number>"
  elseif arg_type == "string" then
    return "<string>" .. arg .. "</string>"
  elseif arg_type == "boolean" then
    return arg and "<true/>" or "<false/>"
  elseif arg_type == "null" then
    return "<null/>"
  elseif arg_type == "table" then
    local index, value = next(arg)
    if index == nil then return "<object></object>" end
    local arg_list = {}
    if index == 1 then
      -- Array
      table.insert(arg_list, "<array>")
      for id, prop in ipairs(arg) do
        table.insert(arg_list, ([[<property id="%d">%s</property>]]):format(id, buildCallFunctionArg(prop)))
      end
      table.insert(arg_list, "</array>")
      return table.concat(arg_list, "")
    else
      -- Object
      table.insert(arg_list, "<object>")
      for id, prop in pairs(arg) do
        table.insert(arg_list, ([[<property id="%s">%s</property>]]):format(id, buildCallFunctionArg(prop)))
      end
      table.insert(arg_list, "</object>")
      return table.concat(arg_list, "")
    end
  else
    assert(false, "Unsupported!")
  end
end


local function buildCallFunctionArgs(args)
  local arg_list = {}
  for _,arg in ipairs(args) do
    table.insert(arg_list, buildCallFunctionArg(arg))
  end
  return #arg_list == 0 and "" or table.concat(arg_list, "")
end


function buildCallFunction(self, ...)
  local args = {...}
  assert(#args > 0)
  local name = args[1]
  table.remove(args, 1)
  local args_str = buildCallFunctionArgs(args)
  local xml_str = ([['<invoke name="%s" returntype="xml"><arguments>%s</arguments></invoke>']]):format(name, args_str)
  return function()
    return self:callFunction(xml_str)
  end
end


function loadCallFunction(self, filename)
  assert(filename)
  local xml_str = ""
  for line in io.lines(filename) do
    xml_str = xml_str .. line
  end
  return function()
    return self:callFunction(xml_str)
  end
end


function setInterval(self, interval)
  interval = interval and interval or DefaultInterval
  self:sendMessage("SwfSetInterval", interval)
  return getret_int()
end


function getInterval(self)
  self:sendMessage("SwfGetInterval")
  return getret_int()
end


-- fluent interface

function withOpen(self, file, async)
  assert(self:open(file, async) == Success, "cannot open " .. file)
  return self
end


function withMove(self, x, y)
  assert(self:move(x, y) == Success)
  return self
end


function withResize(self, width, height)
  assert(self:resize(width, height) == Success)
  return self
end


function withShow(self)
  self:show()
  return self
end


function withHide(self)
  self:hide()
  return self
end


function withLoop(self, on)
  assert(self:setLoop(on) == Success)
  return self
end


function withQuality(self, quality)
  assert(self:setQuality(quality) == Success)
  return self
end


function withCurrentFrame(self, frame)
  assert(self:setCurrentFrame(frame) == Success)
  return self
end


function withVariable(self, name, value)
  assert(self:setVariable(name, value) == Success)
  return self
end


function withFlashVars(self, vars)
  assert(self:setFlashVars(vars) == Success)
  return self
end


function withOptions(self, options)
  if options.x or options.y then assert(self:move(options.x, options.y) == Success) end
  if options.width or options.height then assert(self:resize(options.width, options.height) == Success) end
  if options.visible ~= nil then assert(self:setVisible(options.visible) == Success) end
  if options.loop ~= nil then assert(self:setLoop(options.loop) == Success) end
  if options.quality then assert(self:setQuality(options.quality) == Success) end
  if options.frame then assert(self:setCurrentFrame(options.frame) == Success) end
  if options.vars then assert(self:setFlashVars(options.vars) == Success) end
  return self
end
