require 'lunit'

local builtFunctions = {}

-- mock

local swf = {
  buildCallFunction = function(self, name, args)
    return function()
      table.insert(builtFunctions, {name = name, args = args})
      return true
    end
  end,
  getInterval = function()
    return 33
  end
}

-- target
local nscrm3d = require 'nscrm3d'

module( "nscrm3d_test", package.seeall, lunit.testcase )


function setup()
  m3d = nscrm3d.new(swf)
end


function teardown()
  builtFunctions = {}
end


function test_lm3d_withoutParams()
  local result = pcall(m3d.lm3d, m3d)
  assert_false(result)
end


function test_lm3d_withoutID()
  local result = pcall(m3d.lm3d, m3d, nil, 'test.swf')
  assert_false(result)
end


function test_lm3d_withoutURL()
  local result = pcall(m3d.lm3d, m3d, 1)
  assert_false(result)
end


function test_lm3d_leastParams()
  m3d:lm3d(1, 'test.swf')
  assert_equal(2, #builtFunctions)
  local args = builtFunctions[1].args
  assert_equal('model.load', args.command)
  assert_equal(1, args.id)
  assert_equal('test.swf', args.url)
  local args = builtFunctions[2].args
  assert_equal('model.loaded', args.command)
  assert_equal(1, args.id)
end


function test_lm3d_mostParams()
  m3d:lm3d(1, 'test.swf', 2, 3, 4, 255)
  assert_equal(2, #builtFunctions)
  local args = builtFunctions[1].args
  assert_equal('model.load', args.command)
  assert_equal(1, args.id)
  assert_equal('test.swf', args.url)
  assert_equal(2, args.x)
  assert_equal(-3, args.y)
  assert_equal(4, args.z)
  assert_equal(100, args.alpha)
  local args = builtFunctions[2].args
  assert_equal('model.loaded', args.command)
  assert_equal(1, args.id)
end


function test_cm3d_withoutParams()
  local result = pcall(m3d.cm3d, m3d)
  assert_false(result)
end


function test_cm3d_validParams()
  m3d:cm3d(1)
  assert_equal(1, #builtFunctions)
  local args = builtFunctions[1].args
  assert_equal('model.clear', args.command)
  assert_equal(1, args.id)
end


function test_mm3d_withoutParams()
  local result = pcall(m3d.mm3d, m3d)
  assert_false(result)
end


function test_mm3d_leastParams()
  m3d:mm3d(1)
  assert_equal(1, #builtFunctions)
  local args = builtFunctions[1].args
  assert_equal('model.moveBy', args.command)
  assert_equal(1, args.id)
end


function test_mm3d_mostParams()
  m3d:mm3d(1, 2, 3, 4, 255)
  assert_equal(1, #builtFunctions)
  local args = builtFunctions[1].args
  assert_equal('model.moveBy', args.command)
  assert_equal(1, args.id)
  assert_equal(2, args.x)
  assert_equal(-3, args.y)
  assert_equal(4, args.z)
  assert_equal(100, args.alpha)
end


function test_amm3d_withoutParams()
  local result = pcall(m3d.amm3d, m3d)
  assert_false(result)
end


function test_amm3d_leastParams()
  m3d:amm3d(1)
  assert_equal(1, #builtFunctions)
  local args = builtFunctions[1].args
  assert_equal('model.moveTo', args.command)
  assert_equal(1, args.id)
end


function test_amm3d_mostParams()
  m3d:amm3d(1, 2, 3, 4, 255)
  assert_equal(1, #builtFunctions)
  local args = builtFunctions[1].args
  assert_equal('model.moveTo', args.command)
  assert_equal(1, args.id)
  assert_equal(2, args.x)
  assert_equal(-3, args.y)
  assert_equal(4, args.z)
  assert_equal(100, args.alpha)
end


function test_vm3d_withoutParams()
  local result = pcall(m3d.vm3d, m3d)
  assert_false(result)
end


function test_vm3d_validParams()
  m3d:vm3d(1, true)
  assert_equal(1, #builtFunctions)
  local args = builtFunctions[1].args
  assert_equal('model.visible', args.command)
  assert_equal(1, args.id)
  assert_true(args.visible)
end


function test_rm3d_withoutParams()
  local result = pcall(m3d.rm3d, m3d)
  assert_false(result)
end


function test_rm3d_leastParams()
  m3d:rm3d(1)
  assert_equal(1, #builtFunctions)
  local args = builtFunctions[1].args
  assert_equal('model.rotateBy', args.command)
  assert_equal(1, args.id)
end


function test_rm3d_mostParams()
  m3d:rm3d(1, 2, 3, 4)
  assert_equal(1, #builtFunctions)
  local args = builtFunctions[1].args
  assert_equal('model.rotateBy', args.command)
  assert_equal(1, args.id)
  assert_equal(2, args.x)
  assert_equal(-3, args.y)
  assert_equal(4, args.z)
end


function test_arm3d_withoutParams()
  local result = pcall(m3d.arm3d, m3d)
  assert_false(result)
end


function test_arm3d_leastParams()
  m3d:arm3d(1)
  assert_equal(1, #builtFunctions)
  local args = builtFunctions[1].args
  assert_equal('model.rotateTo', args.command)
  assert_equal(1, args.id)
end


function test_arm3d_mostParams()
  m3d:arm3d(1, 2, 3, 4)
  assert_equal(1, #builtFunctions)
  local args = builtFunctions[1].args
  assert_equal('model.rotateTo', args.command)
  assert_equal(1, args.id)
  assert_equal(2, args.x)
  assert_equal(-3, args.y)
  assert_equal(4, args.z)
end


