package {
	import flash.display.Sprite;
	import flash.events.Event;
	import flash.events.KeyboardEvent;
	import flash.external.ExternalInterface;
	import flash.text.TextField;
	import flash.text.TextFieldAutoSize;
	import ModelEvent;
	
	/**
	 * ...
	 * @author senzogawa
	 */
	public class Main extends Sprite {
		private static const MAX_TRACE_LINES:Number = 20;
		private var _engine:Engine;
		private var _modelController:ModelController;
		private var _motionController:MotionController;
		private var _cameraController:CameraController;
		
		private var _traces:Vector.<String> = new Vector.<String>();
		private var _traceField:TextField;
		private var _shownTrace:Boolean;
		
		public function Main():void {
			if (stage) init();
			else addEventListener(Event.ADDED_TO_STAGE, init);
		}
		
		private function init(e:Event = null):void {
			removeEventListener(Event.ADDED_TO_STAGE, init);
			// entry point
			_engine = new Engine(this);
			_modelController = new ModelController(_engine.view);
			_motionController = new MotionController(_modelController);
			_cameraController = new CameraController(_engine.view);
			
			_modelController.addEventListener(ModelEvent.CLEAR, onModelClear);
			
			initForNScripter();
			if (CONFIG::debug) {
				stage.addEventListener(KeyboardEvent.KEY_DOWN, onKeyDown);
				_engine.renderOn = Engine.RENDER_EVERYFRAME;
			}
			addEventListener(Event.ENTER_FRAME, onEnterFrame);
		}
		
		private function onModelClear(event:ModelEvent):void {
			_motionController.clear(event.id);
		}
		
		private function onEnterFrame(event:Event):void {
			_motionController.onEnterFrame(event);
			_engine.onEnterFrame(event);
		}
		
		private function initForNScripter():void {
			if (!CONFIG::release) return;
			ExternalInterface.addCallback("doCommand", doCommand);
			var self:Main = this;
			Log.tracer = traceToText;
			_traceField = new TextField();
			_traceField.autoSize = TextFieldAutoSize.LEFT;
			_traceField.multiline = true;
		}
		
		private function traceToText(...args):void {
			_traces.push(args.join("\t"));
			if (_traces.length > MAX_TRACE_LINES) {
				_traces.shift();
			}
			if (_shownTrace) {
				_traceField.text = _traces.join("\r");
			}
		}
		
		private function doCommand(args:Object):* {
			if (args == null || args.command == undefined) return null;
			var command:String = args.command;
			var ids:Array = command.split(".");
			try {
				if (ids.length == 1) {
					// Global
					switch(ids[0]) {
					case "trace":
						setTraceVisible(args.value as Boolean);
						return true;
					case "renderOn":
						_engine.renderOn = args.value as Number;
						return true;
					}
					return null;
				}
				else {
					switch(ids[0]) {
					case "model":
						return _modelController.dispatchCommand(ids[1], args);
					case "motion":
						return _motionController.dispatchCommand(ids[1], args);
					case "camera":
						return _cameraController.dispatchCommand(ids[1], args);
					}
				}
			}
			catch (e:Error) {
				Log.e(e.getStackTrace());
			}
			return null;
		}
		
		private function onKeyDown(event:KeyboardEvent):void {
			var char:String = String.fromCharCode(event.charCode);
			if (char == null) return;
			switch(char.toLowerCase()) {
			case "l":
				doCommand( { command:"model.load", id:0, url:CONFIG::testFile } );
				break;
			case "c":
				doCommand( { command:"model.clear", id:0 } );
				break;
			case "v":
				doCommand( { command:"motion.load", id:0, url:CONFIG::testMotion } );
				break;
			case "m":
				doCommand( { command:"motion.load", id:0, url:CONFIG::testMorph, append:true } );
				break;
			case "p":
				doCommand( { command:"motion.play", id:0 } );
				break;
			case "s":
				doCommand( { command:"motion.stop", id:0 } );
				break;
			case "-":
				doCommand( { command:"motion.clear", id:0 } );
				break;
			case "0":
				doCommand( { command:"model.visible", id:0, visible:false } );
				break;
			case "1":
				doCommand( { command:"model.visible", id:0, visible:true } );
				break;
			case "2":
				doCommand( { command:"model.moveBy", id:0, x:40, y:10, z:20, alhpa:-0.1 } );
				break;
			case "3":
				doCommand( { command:"model.moveTo", id:0, x:0, y:0, z:0, alpha:100 } );
				break;
			case "4":
				doCommand( { command:"model.rotateBy", id:0, x:40, y:10, z:20 } );
				break;
			case "5":
				doCommand( { command:"model.rotateTo", id:0, x:0, y:0, z:0 } );
				break;
			}
		}
		
		private function setTraceVisible(visible:Boolean):void {
			if (_traceField == null) return;
			_shownTrace = visible;
			if (_shownTrace) {
				addChild(_traceField);
			}
			else {
				removeChild(_traceField);
			}
		}

	}
	
}