package  {
	import away3dlite.containers.Scene3D;
	import away3dlite.containers.View3D;
	import away3dlite.core.base.Mesh;
	import away3dlite.core.base.Object3D;
	import away3dlite.events.Loader3DEvent;
	import away3dlite.loaders.AbstractParser;
	import away3dlite.loaders.Collada;
	import away3dlite.loaders.Loader3D;
	import away3dlite.loaders.Max3DS;
	import away3dlite.loaders.MD2;
	import away3dlite.loaders.MQO;
	import flash.events.EventDispatcher;
	import net.gamekotan.mmd.PMD;
	import net.gamekotan.mmd.PMDAnimation;

	/**
	 * Dispatched when the model controller will clear.
	 * 
	 * @eventType ModelEvent
	 */
	[Event(name="clear",type="ModelEvent")]

	/**
	 * ...
	 * @author senzogawa
	 */
	public class ModelController extends EventDispatcher {
		private var _view:View3D;
		private var _loaders:Vector.<Loader3D>;
		private var _animations:Vector.<PMDAnimation>;
		private var _loadings:Array;
		
		public function ModelController(view:View3D) {
			this._view = view;
			_loaders = new Vector.<Loader3D>();
			_animations = new Vector.<PMDAnimation>();
			_loadings = new Array();
		}
		
		public function load(id:int, url:String):Boolean {
			var parser:AbstractParser = createParser(url);
			var loader:Loader3D = new Loader3D();
			loader.addEventListener(Loader3DEvent.LOAD_SUCCESS, onLoadSuccess);
			loader.addEventListener(Loader3DEvent.LOAD_ERROR, onLoadError);
			loader.loadGeometry(url, parser);
			_view.scene.addChild(loader);
			return true;
		}
		
		public function clear(id:int):Boolean {
			// TODO: ロード中の消去はキャンセルとして扱う。
			if (_loaders[id] == null) return false;
			dispatchEvent(new ModelEvent(ModelEvent.CLEAR, id));
			var model:Object3D = _loaders[id].handle;
			if(_view.scene.contains(model))
				_view.scene.removeChild(model);
			_loaders[id] = null;
			_animations[id] = null;
			return true;
		}
		
		private function createParser(url:String):AbstractParser {
			if (url.match(/\.pmd$/i)) return new PMD();
			if (url.match(/\.mqo$/i)) return new MQO();
			if (url.match(/\.dae$/i)) return new Collada();
			if (url.match(/\.3ds$/i)) return new Max3DS();
			if (url.match(/\.md2$/i)) return new MD2();
			throw new Error("Unsupported Format:" + url);
		}
		
		private function onLoadSuccess(event:Loader3DEvent):void {
			var args:Object = _loadings.shift();
			var id:int = args.id;
			var loader:Loader3D = event.loader;
			if (loader.parser is PMD) {
				var pmd:PMD = loader.parser as PMD;
				_animations[id] = PMDAnimation.createFromPMD(pmd);
			}
			_loaders[id] = loader;
			moveBy(id, args.x, args.y, args.z, args.alpha);
			loadNext();
		}
		
		private function onLoadError(event:Loader3DEvent):void {
			var args:Object = _loadings.shift();
			Log.e("LoadError:" + args.id);
			_view.scene.removeChild(event.loader);
			loadNext();
		}
		
		private function loadNext():void {
			if (_loadings.length == 0) return;
			var args:Object = _loadings[0];
			load(args.id, args.url);
		}
		
		public function dispatchCommand(command:String, args:Object):* {
			if (args.id < 0) throw new Error("Illegal ID:" + args.id);
			Log.v(args.command);
			switch(command) {
			case "load":
				_loadings.push(args);
				if (_loadings.length > 1) {
					return false;
				}
				return load(args.id, args.url);
			case "clear":
				return clear(args.id);
			case "loaded":
				return isLoaded(args.id);
			case "visible":
				return setVisible(args.id, args.visible);
			case "moveBy":
				return moveBy(args.id, args.x, args.y, args.z, args.alpha);
			case "moveTo":
				return moveTo(args.id, args.x, args.y, args.z, args.alpha);
			case "rotateBy":
				return rotateBy(args.id, args.x, args.y, args.z);
			case "rotateTo":
				return rotateTo(args.id, args.x, args.y, args.z);
			}
			return null;
		}
		
		private function rotateBy(id:int, x:int, y:int, z:int):Boolean {
			var loader:Loader3D = _loaders[id];
			if (loader == null) return false;
			var model:Object3D = getMesh(id);
			model.rotationX -= x;
			model.rotationY -= y;
			model.rotationZ -= z;
			return true;
		}
		
		private function rotateTo(id:int, x:int, y:int, z:int):Boolean {
			var loader:Loader3D = _loaders[id];
			if (loader == null) return false;
			var model:Object3D = getMesh(id);
			model.rotationX = -x;
			model.rotationY = -y;
			model.rotationZ = -z;
			return true;
		}
		
		private function moveBy(id:int, x:int, y:int, z:int, alpha:Number):Boolean {
			var loader:Loader3D = _loaders[id];
			if (loader == null) return false;
			var model:Object3D = getMesh(id);
			model.x -= x;
			model.y -= y;
			model.z -= z;
			// TODO Set Alpha
			return true;
		}
		
		private function moveTo(id:int, x:int, y:int, z:int, alpha:Number):Boolean {
			var loader:Loader3D = _loaders[id];
			if (loader == null) return false;
			var model:Object3D = getMesh(id);
			model.x = -x;
			model.y = -y;
			model.z = -z;
			// TODO Set Alpha
			return true;
		}
		
		private function setVisible(id:int, visible:Boolean):Boolean {
			var scene:Scene3D = _view.scene;
			var loader:Loader3D = _loaders[id];
			if (loader == null) return false;
			var model:Object3D = loader.handle;
			if (!scene.contains(model) && visible) {
				scene.addChild(model);
			}
			else if (scene.contains(model) && !visible) {
				scene.removeChild(model);
			}
			return true;
		}
		
		public function isLoaded(id:int):Boolean {
			return !isLoading(id) && _loaders[id] != null;
		}
		
		public function canAnimate(id:int):Boolean {
			return _animations[id] != null;
		}
		
		public function getMesh(id:int):Mesh {
			return _loaders[id].handle.getChildByName(PMD.MAIN_MESH_NAME) as Mesh;
		}
		
		public function getAnimation(id:int):PMDAnimation {
			return _animations[id];
		}
		
		private function isLoading(id:int):Boolean {
			return _loadings.some(function(item:Object, index:int, array:Array):Boolean {
				return item.id == id;
			});
		}
	}

}