package  {
	import flash.events.Event;
	import net.gamekotan.mmd.MMDAnimator;
	import net.gamekotan.mmd.VMD;
	import net.gamekotan.mmd.VMDEvent;
	/**
	 * ...
	 * @author senzogawa
	 */
	public class MotionController {
		private var _modelController:ModelController;
		private var _animators:Vector.<MMDAnimator> = new Vector.<MMDAnimator>();
		private var _loadings:Array = new Array();
		private var _playings:Vector.<int> = new Vector.<int>();
		
		public function MotionController(modelController:ModelController) {
			this._modelController = modelController;
		}
		
		public function load(id:int, url:String):Boolean {
			if (!_modelController.isLoaded(id)) {
				// TODO モデルのロードを待ってロードする。
				throw new Error("Invalid State:" + id);
			}
			if (!_modelController.canAnimate(id)) {
				// PMD以外はとりあえず非対応。
				throw new Error("Unsupported Animate:" + id);
			}
			var vmd:VMD = new VMD();
			vmd.addEventListener(VMDEvent.LOAD_SUCCESS, onLoadSuccess);
			vmd.addEventListener(VMDEvent.LOAD_ERROR, onLoadError);
			vmd.load(url);
			return true;
		}
		
		private function onLoadSuccess(event:VMDEvent):void {
			var args:Object = _loadings.shift();
			var id:int = args.id;
			var vmd:VMD = event.vmd;
			vmd.parse();
			// TODO appendに対応すること。
			if (_animators.length > id && _animators[id] != null) {
				stop(id);
			}
			_animators[id] = new MMDAnimator(_modelController.getMesh(id), _modelController.getAnimation(id), vmd);
			_animators[id].onRenderedLastFrame = onRenderedLastFrame;
			loadNext();
		}
		
		private function onRenderedLastFrame(animator:MMDAnimator):void {
			if (animator.loop) return;
			var id:int = _animators.indexOf(animator);
			if (id < 0) return;
			stop(id);
		}
		
		private function onLoadError(event:VMDEvent):void {
			var args:Object = _loadings.shift();
			Log.e("LoadError:" + args.id);
			loadNext();
		}
		
		private function loadNext():void {
			if (_loadings.length == 0) return;
			var args:Object = _loadings[0];
			load(args.id, args.url);
		}
		
		public function clear(id:int):void {
			if (id >= _animators.length || _animators[id] == null) return;
			stop(id);
			_animators[id].dispose();
		}
		
		public function play(id:int, frame:int = 1):Boolean {
			if (_animators[id] == null || _animators[id].disposed) return false;
			// TODO フレーム指定に対応すること。
			if (_playings.indexOf(id) >= 0) return true;
			_playings.push(id);
			return true;
		}
		
		public function stop(id:int, frame:int = 1):Boolean {
			if (_animators[id] == null || _animators[id].disposed) return false;
			// TODO フレーム指定に対応すること。
			var index:int = _playings.indexOf(id);
			if (index >= 0) {
				_playings.splice(index, 1);
			}
			return true;
		}
		
		public function dispatchCommand(command:String, args:Object):* {
			if (args.id < 0) throw new Error("Illegal ID:" + args.id);
			switch(command) {
			case "load":
				_loadings.push(args);
				if (_loadings.length > 1) {
					return false;
				}
				return load(args.id, args.url);
			case "clear":
				clear(args.id);
				return true;
			case "play":
				return play(args.id, args.frame);
			case "stop":
				return stop(args.id, args.frame);
			case "loaded":
				return !isLoading(args.id) && _animators[args.id] != null;
			}
			return null;
		}
		
		private function isLoading(id:int):Boolean {
			return _loadings.some(function(item:Object, index:int, array:Array):Boolean {
				return item.id == id;
			});
		}
		
		public function onEnterFrame(event:Event):void {
			for each(var id:int in _playings) {
				_animators[id].render();
			}
		}
	}

}