package net.gamekotan.mmd {
	import away3dlite.core.base.Mesh;
	import away3dlite.core.utils.Debug;
	import flash.events.Event;
	import flash.geom.Matrix;
	import flash.geom.Matrix3D;
	import flash.geom.Vector3D;
	import flash.utils.getTimer;
	/**
	 * ...
	 * @author senzogawa
	 */
	public class MMDAnimator {
		
		private var _mesh:Mesh;
		private var _animation:PMDAnimation;
		private var _vmd:VMD;
		private var _boneMotionLibrary:MMDBoneMotionLibrary;
		private var _morphMotionLibrary:MMDMorphMotionLibrary;
		
		private var _currentFrame:int;
		private var _originalVertices:Vector.<Vector3D> = new Vector.<Vector3D>();
		private var _bufferVertices:Vector.<Number>;
		private var _originalMeshVertices:Vector.<Number>;
		
		private var _disposed:Boolean = false;
		private var _loop:Boolean;
		public var onRenderedLastFrame:Function;
		
		public function MMDAnimator(mesh:Mesh, animation:PMDAnimation, vmd:VMD) {
			_mesh = mesh;
			_animation = animation;
			_vmd = vmd;
			_originalMeshVertices = _mesh.vertices.concat();
			init();
		}
		
		private function init():void {
			var animation:PMDAnimation = _animation;
			// ボーン
			_boneMotionLibrary = new MMDBoneMotionLibrary();
			for each(var vmdMotion:VMDMotion in _vmd.motions) {
				if (!animation.getBoneByName(vmdMotion.name)) continue; // 適用先ボーンが無い
				var boneMotion:MMDBoneMotion = _boneMotionLibrary.addBoneMotion(vmdMotion.name);
				boneMotion.motions.push(vmdMotion);
			}
			_boneMotionLibrary.eachData(function(boneMotion:MMDBoneMotion):void {
				// 昇順ソート
				boneMotion.sort();
				boneMotion.motions.fixed = true;
			});
			// 表情
			_morphMotionLibrary = new MMDMorphMotionLibrary();
			for each(var vmdMorph:VMDMorph in _vmd.morphs) {
				var morphMotion:MMDMorphMotion = _morphMotionLibrary.addMorphMotion(vmdMorph.name);
				morphMotion.morphs.push(vmdMorph);
			}
			_morphMotionLibrary.eachData(function(morphMotion:MMDMorphMotion):void {
				morphMotion.sort();
				morphMotion.morphs.fixed = true;
			});
			// 他
			_currentFrame = 0;
			_loop = _vmd.maxFrame > 0; // 0の場合は単純な表情変更とみなす。
			_originalVertices = animation.vertices;
			_bufferVertices = new Vector.<Number>(_originalVertices.length * 3, true);
		}
		
		
		public function play():void {
			Debug.trace("Play!");
			if (_mesh.hasEventListener(Event.ENTER_FRAME)) return;
			_mesh.addEventListener(Event.ENTER_FRAME, onEnterFrame);
		}
		
		public function stop():void {
			Debug.trace("Stop!");
			if (!_mesh.hasEventListener(Event.ENTER_FRAME)) return;
			_mesh.removeEventListener(Event.ENTER_FRAME, onEnterFrame);
		}
		
		private function updateMotion():void {
			// 表情のリセット
			if (_animation.morphs.length > 0) {
				// 先頭はBase
				_animation.morphs[0].setTo(_originalVertices);
			}
			// モーション更新
			_boneMotionLibrary.eachData(function(boneMotion:MMDBoneMotion):void {
				boneMotion.update(_animation.getBoneByName(boneMotion.name), _currentFrame);
			});
			_morphMotionLibrary.eachData(function(morphMotion:MMDMorphMotion):void {
				morphMotion.update(_animation.getMorphByName(morphMotion.name), _originalVertices, _currentFrame);
			});
			// ボーン行列の更新
			for each(var bone:PMDBone in _animation.bones) {
				bone.updateTransform();
			}
			// IKの更新
			for each(var ik:PMDIK in _animation.iks) {
				ik.update();
			}
		}
		
		private function skinning():void {
			for each(var bone:PMDBone in _animation.bones) {
				bone.updateSkinning();
			}
			// ボーンに応じた頂点の更新
			var position:Vector3D;
			var originalVertices:Vector.<Vector3D> = _originalVertices;
			var vertices:Vector.<Number> = _bufferVertices;
			var length:int = originalVertices.length;
			var v:int = 0;
			for (var i:int = 0; i < length; i++) {
				var skin:PMDSkin = _animation.skins[i];
				if (skin.weight == 0.0) {
					position = skin.bones[1].skinning(originalVertices[i]);
				}
				else if (skin.weight >= 0.9999) {
					position = skin.bones[0].skinning(originalVertices[i]);
				}
				else {
					var transform:Matrix3D = skin.getLerpMatrix();
					position = transform.transformVector(originalVertices[i]);
				}
				vertices[v++] = -position.x;
				vertices[v++] = -position.y;
				vertices[v++] = -position.z;
			}
			// 頂点に応じたメッシュ頂点配列の更新
			i = 0;
			var meshVertices:Vector.<Number> = _mesh.vertices;
			var faces:Vector.<PMDFace> = _animation.faces;
			for each(var face:PMDFace in faces) {
				// 今のところこれが一番速い。
				meshVertices[i++] = vertices[face.v0];
				meshVertices[i++] = vertices[face.v1];
				meshVertices[i++] = vertices[face.v2];
				meshVertices[i++] = vertices[face.v3];
				meshVertices[i++] = vertices[face.v4];
				meshVertices[i++] = vertices[face.v5];
				meshVertices[i++] = vertices[face.v6];
				meshVertices[i++] = vertices[face.v7];
				meshVertices[i++] = vertices[face.v8];
			}
		}
		
		public function render():void {
			Debug.trace(_currentFrame);
			updateMotion();
			skinning();
			_currentFrame++;
			if (_currentFrame > _vmd.maxFrame) {
				if (onRenderedLastFrame != null) onRenderedLastFrame(this);
				_currentFrame = 0;
			}
		}
		
		public function get loop():Boolean {
			return _loop;
		}
		
		public function set loop(value:Boolean):void {
			loop = _loop;
		}
		
		public function dispose():void {
			if (_disposed) return;
			_disposed = true;
			stop();
			var vmd:VMD = new VMD();
			vmd.maxFrame = 0;
			vmd.motions = new Vector.<VMDMotion>();
			vmd.morphs = new Vector.<VMDMorph>();
			vmd.name = "Base";
			_vmd = vmd;
			init();
			_animation.morphs[0].setTo(_originalVertices);
			var i:int = _originalMeshVertices.length;
			while (i) {
				i--;
				_mesh.vertices[i] = _originalMeshVertices[i];
			}
		}
		
		public function get disposed():Boolean {
			return _disposed;
		}
		
		private function onEnterFrame(event:Event):void {
			Debug.trace(_currentFrame);
			updateMotion();
			skinning();
			_currentFrame++;
			if (_currentFrame > _vmd.maxFrame) {
				if (_vmd.maxFrame == 0) {
					stop();
				}
				_currentFrame = 0;
			}
		}
	}

}