package net.gamekotan.mmd {
	import flash.geom.Matrix3D;
	import flash.geom.Vector3D;
	import net.gamekotan.util.Quarternion;
	/**
	 * ...
	 * @author senzogawa
	 */
	public class MMDBoneMotion {
		
		public var name:String;
		public var motions:Vector.<VMDMotion>;
		
		public function MMDBoneMotion(name:String) {
			this.name = name;
			motions = new Vector.<VMDMotion>();
		}
		
		public function sort():void {
			// Sort Ascendant
			motions.sort(function(one:VMDMotion, another:VMDMotion):Number {
				return one.frame - another.frame;
			});
		}
		
		public function update(bone:PMDBone, frame:int):void {
			var length:int = motions.length;
			var last:int = length - 1;
			
			var current:int;
			var next:int;
			if (frame >= motions[last].frame) {
				current = next = last;
			}
			else if (frame <= motions[0].frame) {
				current = next = 0;
			}
			else {
				next = last;
				for (var i:int = 0; i < length; i++) {
					if (frame <= motions[i].frame) {
						next = i;
						break;
					}
				}
				current = Math.max(0, next - 1);
			}
			var currentMotion:VMDMotion = motions[current];
			var nextMotion:VMDMotion = motions[next];
			var translation:Vector3D;
			var rotation:Quarternion;
			if (current != next) {
				var lerp:Number = (frame - currentMotion.frame) / (nextMotion.frame - currentMotion.frame);
				translation = lerpVector3D(currentMotion.translation, nextMotion.translation, lerp);
				rotation = currentMotion.rotation.slerp(nextMotion.rotation, lerp);
			}
			else {
				translation = currentMotion.translation.clone();
				rotation = currentMotion.rotation.clone();
			}
			rotation.normalize();
			bone.translation = translation;
			bone.rotation = rotation;
		}
		
		private static function lerpVector3D(src:Vector3D, dst:Vector3D, lerp:Number):Vector3D {
			var t:Number = 1.0 - lerp;
			return new Vector3D(
				src.x * t + dst.x * lerp,
				src.y * t + dst.y * lerp,
				src.z * t + dst.z * lerp);
		}
		
	}

}