package net.gamekotan.mmd {
	import away3dlite.arcane;
	import away3dlite.containers.*;
	import away3dlite.core.base.*;
	import away3dlite.core.utils.Debug;
	import away3dlite.loaders.AbstractParser;
	import away3dlite.loaders.data.*;
	import away3dlite.loaders.utils.GeometryLibrary;
	
	import flash.geom.*;
	import flash.utils.*;
	use namespace arcane;
	
	/**
	 * ...
	 * @author senzogawa
	 */
	public class PMD extends AbstractParser {
		
		public static var MAIN_MESH_NAME:String = "main_mesh";
		
		private var _pmd:ByteArray;
		
		private var _vertexChunk:PMDVertexChunk;
		private var _faceChunk:PMDFaceChunk;
		private var _boneChunk:PMDBoneChunk;
		private var _ikChunk:PMDIKChunk;
		private var _morphChunk:PMDMorphChunk;
		
		/** @private */
        arcane override function prepareData(data:*):void {
			_pmd = data;
			_pmd.endian = Endian.LITTLE_ENDIAN;
		}
		
		/** @private */
        arcane override function parseNext():void {
			var pmd:ByteArray = _pmd;
			
			if (pmd.readMultiByte(3, charset) != "Pmd") {
				// The data is not PMD!
				notifyError();
				return;
			}
			// Header
			var pmdVer:Number = pmd.readFloat();
			var modelName:String = pmd.readMultiByte(20, charset);
			var modelComment:String = pmd.readMultiByte(256, charset);
			Debug.trace(modelName);
			Debug.trace(modelComment);
			
			// Chunks
			_vertexChunk = parseVertex(pmd);			
			_faceChunk = parseFace(pmd);
			parseMaterial(pmd, _faceChunk.faces);
			_boneChunk = parseBone(pmd);
			_ikChunk = parseIK(pmd);
			_morphChunk = parseMorph(pmd);
			
			// Rests are unknown.
			Debug.trace(pmd.bytesAvailable);
			
			// Build
			var geometryData:GeometryData = _geometryLibrary.addGeometry(modelName);
			_vertexChunk.pushToGeometry(geometryData);
			_faceChunk.pushToGeometry(geometryData);
			
			buildMaterials();
			buildMeshes(geometryData);
			
			notifySuccess();
		}
		
		private function parseMorph(pmd:ByteArray):PMDMorphChunk {
			// Morphs
			var count:int = pmd.readShort();
			Debug.trace("Morphs " + count);
			var chunk:PMDMorphChunk = new PMDMorphChunk(count);
			var morphs:Vector.<PMDMorphData> = chunk.morphs;
			for (var i:int = 0; i < count; i++) {
				var data:PMDMorphData = new PMDMorphData(pmd.readMultiByte(20, charset), pmd.readInt(), pmd.readByte());				
				var vcount:int = data.vertices.length;
				var vertices:Vector.<PMDMorphVertex> = data.vertices;
				for (var j:int = 0; j < vcount; j++) {
					vertices[j] = new PMDMorphVertex(
						pmd.readInt(),
						new Vector3D(pmd.readFloat() * scaling, pmd.readFloat() * scaling, pmd.readFloat() * scaling)
					);
				}
				morphs[i] = data;
			}
			return chunk;
		}
		
		private function parseIK(pmd:ByteArray):PMDIKChunk {
			// IK
			var count:int = pmd.readShort();
			Debug.trace("IK " + count);
			var chunk:PMDIKChunk = new PMDIKChunk(count);
			var iks:Vector.<PMDIKData> = chunk.iks;
			for (var i:int = 0; i < count; i++) {
				var data:PMDIKData = new PMDIKData(
					pmd.readShort(),
					pmd.readShort(),
					pmd.readByte(),
					pmd.readUnsignedShort(),
					pmd.readFloat()
					);
				var jcount:int = data.links.length;
				for (var j:int = 0; j < jcount; j++) {
					data.links[j] = pmd.readUnsignedShort();
				}
				iks[i] = data;
			}
			return chunk;
		}
		
		private function parseBone(pmd:ByteArray):PMDBoneChunk {
			// Bones
			var count:int = pmd.readShort();
			Debug.trace("Bones " + count);
			var chunk:PMDBoneChunk = new PMDBoneChunk(count);
			var bones:Vector.<PMDBoneData> = chunk.bones;
			for (var i:int = 0; i < count; i++) {
				bones[i] = new PMDBoneData(
					pmd.readMultiByte(20, charset),
					pmd.readShort(),
					pmd.readShort(),
					pmd.readByte(),
					pmd.readShort(),
					new Vector3D(pmd.readFloat(), pmd.readFloat(), pmd.readFloat())
					);
			}
			return chunk;
		}
		
		private function parseMaterial(pmd:ByteArray, faces:Vector.<PMDFaceData>):void {
			// Materials
			var count:int = pmd.readInt();
			var foffset:int = 0;
			for (var i:int = 0; i < count; i++) {
				var materialData:MaterialData = _materialLibrary.addMaterial("mat" + i);
				_materialNames.push(materialData.name);
				
				var diffuseColor:uint = rgbToUInt(pmd.readFloat(), pmd.readFloat(), pmd.readFloat());
				var a:Number = pmd.readFloat();
				
				var shininess:Number = pmd.readFloat();
				var specularColor:uint = rgbToUInt(pmd.readFloat(), pmd.readFloat(), pmd.readFloat());
				var ambientColor:uint = rgbToUInt(pmd.readFloat(), pmd.readFloat(), pmd.readFloat());
				skipBytes(pmd, 2);
				
				var fcount:int = pmd.readInt() / 3;
				for (var j:int = 0; j < fcount; j++) {
					faces[j + foffset].materialData = materialData;
				}
				foffset += fcount;
				
				var tex:String = pmd.readMultiByte(20, charset);
				// TODO Implement SphereMap
				if (tex.length == 0 || tex.match(/^\s+$/) || tex.match(/\.(sph|spa)$/)) {
					Debug.trace("Color: " + diffuseColor + "," + a);
					materialData.diffuseColor = diffuseColor;
					materialData.alpha = a;
					
					materialData.shininess = shininess;
					materialData.specularColor = specularColor;
					materialData.ambientColor = ambientColor;
					
					materialData.materialType = shading ? MaterialData.SHADING_MATERIAL : MaterialData.COLOR_MATERIAL;
				}
				else {
					Debug.trace("Texture: " + tex);
					materialData.materialType = MaterialData.TEXTURE_MATERIAL;
					if (tex.match(/\.(tga|bmp)$/i)) {
						tex = tex.replace(/\.(tga|bmp)$/i, ".png");
						Debug.trace("TGA or BMP must be replaced to PNG " + tex);
					}
					materialData.textureFileName = tex;
				}
			}
		}
		
		private static function rgbToUInt(r:Number, g:Number, b:Number):uint {
			r *= 255;
			g *= 255;
			b *= 255;
			return (uint(r) << 16) | (uint(g) << 8) | uint(b);
		}
		
		private function parseFace(pmd:ByteArray):PMDFaceChunk {
			// Faces
			var count:int = pmd.readInt() / 3;
			Debug.trace("Faces " + count);
			var chunk:PMDFaceChunk = new PMDFaceChunk(count);
			var faces:Vector.<PMDFaceData> = chunk.faces;
			for (var i:int = 0; i < count; i++) {
				faces[i] = new PMDFaceData(pmd.readUnsignedShort(), pmd.readUnsignedShort(), pmd.readUnsignedShort());
			}
			return chunk;
		}
		
		private function parseVertex(pmd:ByteArray):PMDVertexChunk {
			// Vertices
			var count:int = pmd.readInt();
			Debug.trace("Vertices " + count);
			var chunk:PMDVertexChunk = new PMDVertexChunk(count);
			var vertices:Vector.<PMDVertexData> = chunk.vertices;
			for (var i:int = 0; i < count; i++) {
				vertices[i] = new PMDVertexData(
					new Vector3D(pmd.readFloat() * scaling, pmd.readFloat() * scaling, pmd.readFloat() * scaling),
					skipBytes(pmd, 12),
					new UV(pmd.readFloat(), pmd.readFloat()),
					[pmd.readUnsignedShort(), pmd.readUnsignedShort()],
					pmd.readByte(),
					pmd.readByte()
					);
			}
			Debug.trace("Vertices[0].vertex " + vertices[0].vertex);
			return chunk;
		}
		
		private function skipBytes(pmd:ByteArray, length:int):Object {
			pmd.readBytes(_dummy, 0, length);
			return null;
		}
		
		// The dummy to skip reading some bytes.
		private var _dummy:ByteArray = new ByteArray();
		
		// The names of materialData for getting from the library.
		private var _materialNames:Array = new Array();
		
		private function buildMeshes(geometryData:GeometryData):void {
			var mesh:Mesh = new Mesh();
			var materialData:MaterialData;
			var i:int = 0;
			var i0:int;
			var i1:int;
			var i2:int;
			var vertices:Vector.<Number> = geometryData.vertices;
			var uvtData:Vector.<Number> = geometryData.uvtData;
			
			for each(var faceData:FaceData in geometryData.faces) {
				// set face materials
				materialData = faceData.materialData;
				mesh._faceMaterials.push(materialData.material);
				
				// Both vertices and uv are ordered with reversed by the length of each face in PMD.
				// So we need to replace 0,1,2 to 2,1,0.
				
				// set vertices
				i0 = faceData.v0*3;
				i1 = faceData.v1*3;
				i2 = faceData.v2*3;
				mesh._vertices.push(vertices[i0], vertices[i0+1], vertices[i0+2]);
				//buildSkinVertices(_geometryData, _faceData.v2, mesh._vertices);
				mesh._vertices.push(vertices[i1], vertices[i1+1], vertices[i1+2]);
				//buildSkinVertices(_geometryData, _faceData.v1, mesh._vertices);
				mesh._vertices.push(vertices[i2], vertices[i2+1], vertices[i2+2]);
				//buildSkinVertices(_geometryData, _faceData.v0, mesh._vertices);
				
				// set uvData
				i0 = faceData.uv0*3;
				i1 = faceData.uv1*3;
				i2 = faceData.uv2*3;
				mesh._uvtData.push(uvtData[i0], uvtData[i0+1], uvtData[i0+2], uvtData[i1], uvtData[i1+1], uvtData[i1+2], uvtData[i2], uvtData[i2+1], uvtData[i2+2]);
				
				// set indices
				mesh._indices.push(i + 2, i + 1, i);
				i += 3;
			}
			mesh.buildFaces();
			mesh.type = ".pmd";
			mesh.name = MAIN_MESH_NAME;
			_container.addChild(mesh);
			
			// Need to use the global material of meshes.
			//for each(var matName:String in _materialNames) {
				//var mat:MaterialData = _materialLibrary.getMaterial(matName);
				//mat.meshes.push(mesh);
			//}
			
			// For setting the material to faces when loaded texture.
			for each(var face:Face in mesh.faces) {
				geometryData.faces[face.index].materialData.faces.push(face);
			}
		}
		
		/* For Animation */
		
		public function getSkinCount():int {
			return _vertexChunk.vertices.length;
		}

		public function getFaceCount():int {
			return _faceChunk.faces.length;
		}
		
		public function getBoneCount():int {
			return _boneChunk.bones.length;
		}
		
		public function getIKCount():int {
			return _ikChunk.iks.length;
		}
		
		public function getMorphCount():int {
			return _morphChunk.morphs.length;
		}
		
		public function eachVertexSkinData(callback:Function):void {
			_vertexChunk.vertices.forEach(function(vertex:PMDVertexData, index:int, vector:Vector.<PMDVertexData>):void {
				callback(index, vertex.vertex, vertex.bones, vertex.weight, vertex.edge);
			});
		}
		
		public function eachFaceData(callback:Function):void {
			_faceChunk.faces.forEach(function(face:PMDFaceData, index:int, vector:Vector.<PMDFaceData>):void {
				callback(index, face.idx0, face.idx1, face.idx2);
			});
		}
		
		public function eachBoneData(callback:Function):void {
			_boneChunk.bones.forEach(function(bone:PMDBoneData, index:int, vector:Vector.<PMDBoneData>):void {
				callback(index, bone.name, bone.parent, bone.child, bone.type, bone.ik, bone.position);
			});
		}
		
		public function eachIKData(callback:Function):void {
			_ikChunk.iks.forEach(function(ik:PMDIKData, index:int, vector:Vector.<PMDIKData>):void {
				callback(index, ik.bone, ik.target, ik.links, ik.loop, ik.limitRadian);
			});
		}
		
		public function eachMorphData(callback:Function):void {
			_morphChunk.morphs.forEach(function(morph:PMDMorphData, index:int, vector:Vector.<PMDMorphData>):void {
				callback(index, morph.name, morph.verticesToVectors());
			});
		}
		
		public function get materialNames():Array {
			return _materialNames;
		}
		
		/* Generic */
		
		public var charset:String = "shift_jis";
        
    	/**
    	 * Controls the use of shading materials when color textures are encountered. Defaults to false.
    	 */
        public var shading:Boolean = false;
        
    	/**
    	 * A scaling factor for all geometry in the model. Defaults to 1.
    	 */
        public var scaling:Number = 1;
        
    	/**
    	 * Controls the automatic centering of geometry data in the model, improving culling and the accuracy of bounding dimension values.
    	 */
        public var centerMeshes:Boolean;
        
		public function PMD() {
			super();
			
			_container = new ObjectContainer3D();
			_container.name = "pmd";
			
			_container.materialLibrary = _materialLibrary;
			_container.geometryLibrary = _geometryLibrary;
			
			binary = true;
		}
		
	}
	
}

import away3dlite.loaders.data.*;
import flash.geom.*;

class UV {
	public var u:Number;
	public var v:Number;
	public function UV(u:Number, v:Number) {
		this.u = u;
		this.v = v;
	}
}

class PMDVertexData {
	public var vertex:Vector3D;
	public var uvt:UV;
	public var bones:Vector.<int>;
	public var weight:int;
	public var edge:int;
	public function PMDVertexData(vertex:Vector3D, normal:*, uvt:UV, bones:Array, weight:int, edge:int) {
		this.vertex = vertex;
		this.uvt = uvt;
		this.bones = new Vector.<int>(2, true);
		this.bones[0] = bones[0];
		this.bones[1] = bones[1];
		this.weight = weight;
		this.edge = edge;
	}
}

class PMDVertexChunk {
	public var vertices:Vector.<PMDVertexData> = new Vector.<PMDVertexData>();
	public function PMDVertexChunk(count:int) {
		vertices = new Vector.<PMDVertexData>(count, true);
	}
	public function pushToGeometry(geometryData:GeometryData):void {
		var gmtVertices:Vector.<Number> = geometryData.vertices;
		var uvtData:Vector.<Number> = geometryData.uvtData;
		var indices:Vector.<int> = geometryData.indices;
		vertices.forEach(function(data:PMDVertexData, index:int, vector:Vector.<PMDVertexData>):void {
			var vtx:Vector3D = data.vertex;
			var uvt:UV = data.uvt;
			gmtVertices.push(-vtx.x, -vtx.y, -vtx.z);
			uvtData.push(uvt.u, uvt.v, 1);
			indices.push(index);
		});
	}
}

class PMDFaceData {
	public var idx0:int;
	public var idx1:int;
	public var idx2:int;
	public var materialData:MaterialData;
	public function PMDFaceData(idx0:int, idx1:int, idx2:int) {
		this.idx0 = idx0;
		this.idx1 = idx1;
		this.idx2 = idx2;
	}
}

class PMDFaceChunk {
	public var faces:Vector.<PMDFaceData>;
	public function PMDFaceChunk(count:int) {
		faces = new Vector.<PMDFaceData>(count, true);
	}
	public function pushToGeometry(geometryData:GeometryData):void {
		var gmtFaces:Array = geometryData.faces;
		faces.forEach(function(data:PMDFaceData, index:int, vector:Vector.<PMDFaceData>):void {
			var face:FaceData = new FaceData();
			face.v0 = face.uv0 = data.idx0;
			face.v1 = face.uv1 = data.idx1;
			face.v2 = face.uv2 = data.idx2;
			face.materialData = data.materialData;
			face.visible = true;
			gmtFaces.push(face);
		});
	}
}

class PMDBoneData {
	public static const PARENT_NONE:int = -1;
	public var name:String;
	public var parent:int;
	public var child:int;
	public var type:int;
	public var ik:int;
	public var position:Vector3D;
	public function PMDBoneData(name:String, parent:int, child:int, type:int, ik:int, position:Vector3D) {
		this.name = name;
		this.parent = parent;
		this.child = child;
		this.type = type;
		this.ik = ik;
		this.position = position;
	}
}

class PMDBoneChunk {
	public var bones:Vector.<PMDBoneData>;
	public function PMDBoneChunk(count:int) {
		bones = new Vector.<PMDBoneData>(count, true);
	}
}

class PMDIKData {
	public var bone:int;
	public var target:int;
	public var loop:int;
	public var limitRadian:Number;
	public var links:Vector.<int>;
	public function PMDIKData(bone:int, target:int, linkCount:int, loop:int, limitRadian:Number) {
		this.bone = bone;
		this.target = target;
		this.links = new Vector.<int>(linkCount, true);
		this.loop = loop;
		this.limitRadian = limitRadian;
	}
}

class PMDIKChunk {
	public var iks:Vector.<PMDIKData>;
	public function PMDIKChunk(count:int) {
		iks = new Vector.<PMDIKData>(count, true);
	}
}

class PMDMorphVertex {
	public var index:int;
	public var vertex:Vector3D;
	public function PMDMorphVertex(index:int, vertex:Vector3D) {
		this.index = index;
		this.vertex = vertex;
	}
	public function toVertexWithIndex():Vector3D {
		var cloned:Vector3D = vertex.clone();
		cloned.w = index;
		return cloned;
	}
}

class PMDMorphData {
	public var name:String;
	public var type:int;
	public var vertices:Vector.<PMDMorphVertex>;
	public function PMDMorphData(name:String, vertexCount:int, type:int) {
		this.name = name;
		this.vertices = new Vector.<PMDMorphVertex>(vertexCount, true);
		this.type = type;
	}
	public function verticesToVectors():Vector.<Vector3D> {
		var vectors:Vector.<Vector3D> = new Vector.<Vector3D>(vertices.length, true);
		var i:int = 0;
		for each(var vertex:PMDMorphVertex in vertices) {
			vectors[i++] = vertex.toVertexWithIndex();
		}
		return vectors;
	}
}

class PMDMorphChunk {
	public var morphs:Vector.<PMDMorphData>;
	public function PMDMorphChunk(count:int) {
		morphs = new Vector.<PMDMorphData>(count, true);
	}
}