package net.gamekotan.mmd {
	import flash.geom.Matrix3D;
	import flash.geom.Vector3D;
	import flash.utils.Dictionary;
	/**
	 * ...
	 * @author senzogawa
	 */
	public class PMDAnimation {
		
		public var vertices:Vector.<Vector3D>;
		public var skins:Vector.<PMDSkin>;
		public var faces:Vector.<PMDFace>;
		public var bones:Vector.<PMDBone>;
		public var iks:Vector.<PMDIK>;
		public var morphs:Vector.<PMDMorph>;
		public var materialNames:Array;
		
		public var transformUser:Matrix3D = new Matrix3D();
		
		private var _boneIndices:Dictionary;
		private var _morphIndices:Dictionary;
		
		public function PMDAnimation(vertexCount:int, faceCount:int, boneCount:int, ikCount:int, morphCount:int) {
			vertices = new Vector.<Vector3D>(vertexCount, true);
			skins = new Vector.<PMDSkin>(vertexCount, true);
			faces = new Vector.<PMDFace>(faceCount, true);
			bones = new Vector.<PMDBone>(boneCount, true);
			for (var i:int = 0; i < boneCount; i++) {
				bones[i] = new PMDBone();
			}
			iks = new Vector.<PMDIK>(ikCount, true);
			morphs = new Vector.<PMDMorph>(morphCount, true);
		}
		
		public function recalcBonesOffset():void {
			for each(var bone:PMDBone in bones) {
				bone.recalcOffset();
			}
		}
		
		public function getBoneByName(name:String):PMDBone {
			if (_boneIndices == null) {
				_boneIndices = createBoneIndices();
			}
			if (_boneIndices[name] != undefined) return bones[_boneIndices[name]];
			return null;
		}
		
		private function createBoneIndices():Dictionary {
			var indices:Dictionary = new Dictionary();
			for (var i:int = 0; i < bones.length; i++) {
				indices[bones[i].name] = i;
			}
			return indices;
		}
		
		public function getMorphByName(name:String):PMDMorph {
			if (_morphIndices == null) {
				_morphIndices = createMorphIndices();
			}
			if (_morphIndices[name] != undefined) return morphs[_morphIndices[name]];
			return null;
		}
		
		private function createMorphIndices():Dictionary {
			var indices:Dictionary = new Dictionary();
			for (var i:int = 0; i < morphs.length; i++) {
				indices[morphs[i].name] = i;
			}
			return indices;
		}
		
		public static function createFromPMD(pmd:PMD):PMDAnimation {
			var animation:PMDAnimation = new PMDAnimation(
				pmd.getSkinCount(), pmd.getFaceCount(), pmd.getBoneCount(), pmd.getIKCount(), pmd.getMorphCount());
			// Bone
			var bones:Vector.<PMDBone> = animation.bones;
			pmd.eachBoneData(function(index:int, name:String, parentBone:int, childBone:int, type:int, ik:int, position:Vector3D):void {
				var bone:PMDBone = bones[index];
				bone.init(name, position);
				if (parentBone >= 0) {
					bone.parent = bones[parentBone];
				}
				if (childBone >= 0) {
					bone.child = bones[childBone];
				}
				bone.transformUser = animation.transformUser;
			});
			// 親子の位置がすべて代入されてからoffsetを計算する必要がある。
			animation.recalcBonesOffset();
			// IK
			var iks:Vector.<PMDIK> = animation.iks;
			pmd.eachIKData(function(index:int, boneIndex:int, targetIndex:int, linkIndices:Vector.<int>, loop:int, limitRadian:Number):void {
				var ik:PMDIK = new PMDIK(bones[boneIndex], bones[targetIndex], loop, linkIndices.length, limitRadian);
				ik.linkBone(linkIndices, bones);
				iks[index] = ik;
			});
			iks.sort(PMDIK.compare);
			// Vertex, Skin
			var vertices:Vector.<Vector3D> = animation.vertices;
			var skins:Vector.<PMDSkin> = animation.skins;
			pmd.eachVertexSkinData(function(index:int, vertex:Vector3D, boneIndices:Vector.<int>, weight:int, edge:int):void {
				var skin:PMDSkin = new PMDSkin(boneIndices.length, weight, edge);
				skin.linkBone(boneIndices, bones);
				vertices[index] = vertex;
				skins[index] = skin;
			});
			// Face
			var faces:Vector.<PMDFace> = animation.faces;
			pmd.eachFaceData(function(index:int, idx0:int, idx1:int, idx2:int):void {
				faces[index] = new PMDFace(idx0, idx1, idx2);
			});
			// Morph
			var morphs:Vector.<PMDMorph> = animation.morphs;
			pmd.eachMorphData(function(index:int, name:String, verticesWithIndex:Vector.<Vector3D>):void {
				morphs[index] = new PMDMorph(name, verticesWithIndex);
				if (index > 0) {
					morphs[index].verticesToWorld(morphs[0]);
				}
			});
			animation.materialNames = pmd.materialNames;
			return animation;
		}
	}

}