package net.gamekotan.mmd {
	import flash.geom.Matrix3D;
	import flash.geom.Orientation3D;
	import flash.geom.Vector3D;
	import net.gamekotan.util.Quarternion;
	/**
	 * ...
	 * @author senzogawa
	 */
	public class PMDBone {
		
		public static const DEFAULT_TRANSLATION:Vector3D = new Vector3D();
		public static const DEFAULT_ROTATION:Quarternion = new Quarternion();
		
		public var transformUser:Matrix3D;
		
		public var name:String;
		
		private var _position:Vector3D;
		private var _offset:Vector3D;
		private var _transformToOrigin:Matrix3D;
		private var _transformCurrent:Matrix3D;
		private var _transformSkinning:Matrix3D;
		
		public var parent:PMDBone;
		public var child:PMDBone;
		
		public var translation:Vector3D = DEFAULT_TRANSLATION;
		public var rotation:Quarternion = DEFAULT_ROTATION;
		
		public var shouldLimitAngle:Boolean = false;
		
		public function init(name:String, position:Vector3D):void {
			this.name = name;
			_position = position.clone();
			_transformToOrigin = new Matrix3D();
			_transformToOrigin.appendTranslation( -_position.x, -_position.y, -_position.z);
			_transformCurrent = _transformToOrigin.clone();
			_transformCurrent.invert();
		}
		
		public function recalcOffset():void {
			if (parent != null) {
				_offset = _position.subtract(parent._position);
			}
			else {
				_offset = _position.clone();
			}
		}
		
		public function updateTransform():void {
			_transformCurrent = rotation.toMatrix3D();
			_transformCurrent.appendTranslation(translation.x + _offset.x, translation.y + _offset.y, translation.z + _offset.z);
			if (parent != null) {
				_transformCurrent.append(parent._transformCurrent);
			}
		}
		
		public function getTransformPosition(position:Vector3D):void {
			var translation:Vector3D = _transformCurrent.decompose(Orientation3D.QUATERNION)[0];
			position.x = translation.x;
			position.y = translation.y;
			position.z = translation.z;
			position.w = 1;
		}
		
		public function getInverseTransform(matrix:Matrix3D):void {
			matrix.rawData = _transformCurrent.rawData;
			matrix.invert();
		}
		
		public function updateSkinning():void {
			_transformSkinning = _transformToOrigin.clone();
			_transformSkinning.append(_transformCurrent);
			_transformSkinning.append(transformUser);
			//trace(name, _transformCurrent.rawData.toString());
		}
		
		public function skinning(position:Vector3D):Vector3D {
			return _transformSkinning.transformVector(position);
		}
		
		public function getLerpMatrix(another:PMDBone, weight:int):Matrix3D {
			// パーセント指定なので、そのまま渡してよい。
			return Matrix3D.interpolate(_transformSkinning, another._transformSkinning, weight);
		}
	}

}