package net.gamekotan.mmd {
	import flash.geom.Matrix3D;
	import flash.geom.Orientation3D;
	import flash.geom.Vector3D;
	import net.gamekotan.util.Quarternion;
	/**
	 * ...
	 * @author senzogawa
	 */
	public class PMDIK {
		
		private var _tip:PMDBone;
		private var _target:PMDBone;
		
		private var _loop:int;
		private var _links:Vector.<PMDBone>;
		private var _limitAngle:Number;
		
		private var _sortValue:int;
		
		public function PMDIK(tip:PMDBone, target:PMDBone, loop:int, linkCount:int, limitRadian:Number) {
			_tip = tip;
			_target = target;
			_loop = loop;
			_links = new Vector.<PMDBone>(linkCount, true);
			_limitAngle = limitRadian * Math.PI;
		}
		
		public function linkBone(linkIndices:Vector.<int>, bones:Vector.<PMDBone>):void {
			_sortValue = linkIndices[0];
			for (var i:int = 0; i < linkIndices.length; i++) {
				var bone:PMDBone = bones[linkIndices[i]];
				bone.shouldLimitAngle = bone.name == "左ひざ" || bone.name == "右ひざ";
				_links[i] = bone;
				//trace(i,linkIndices[i],_links[i].name);
			}
		}
		
		public static function compare(one:PMDIK, another:PMDIK):int {
			return one._sortValue - another._sortValue;
		}
		
		public function update():void {
			var originalPosition:Vector3D = new Vector3D();
			var targetPosition:Vector3D = new Vector3D();
			var tipPosition:Vector3D = new Vector3D();
			var diff:Vector3D;
			var rotAxis:Vector3D;
			var rotQuat:Quarternion = new Quarternion();
			var inverseTransform:Matrix3D = new Matrix3D();
			
			_tip.getTransformPosition(originalPosition);
			//trace("Original",originalPosition);
			updateLinksAndTarget(_links.length - 1);
			for (var i:int = 0; i < _loop; i++) {
				for (var linkIndex:int = 0; linkIndex < _links.length; linkIndex++) {
					var link:PMDBone = _links[linkIndex];
					_target.getTransformPosition(targetPosition);
					//trace(link.name,targetPosition);
					// World to Local
					link.getInverseTransform(inverseTransform);
					
					targetPosition = inverseTransform.transformVector(targetPosition);
					tipPosition = inverseTransform.transformVector(originalPosition);
					
					diff = targetPosition.subtract(tipPosition);
					if (diff.dotProduct(diff) < 0.0000001) {
						return;
					}
					
					targetPosition.normalize();
					tipPosition.normalize();
					
					var angle:Number = Math.acos(targetPosition.dotProduct(tipPosition));
					
					if (0.00000001 < Math.abs(angle)) {
						if (angle < -_limitAngle) angle = -_limitAngle;
						else if (_limitAngle < angle) angle = _limitAngle;
						
						// 回転軸
						rotAxis = targetPosition.crossProduct(tipPosition);
						if (rotAxis.dotProduct(rotAxis) < 0.0000001) continue;
						rotAxis.normalize();
						
						// 関節回転量の補正
						if (Math.abs(angle) < 0.0001) {
							rotQuat.x = rotQuat.y = rotQuat.z = 0;
							rotQuat.w = 1;
						}
						else {
							var tempAngle:Number = angle * 0.5;
							var sinTemp:Number = Math.sin(tempAngle);
							rotQuat.x = rotAxis.x * sinTemp;
							rotQuat.y = rotAxis.y * sinTemp;
							rotQuat.z = rotAxis.z * sinTemp;
							rotQuat.w = Math.cos(tempAngle);
						}
						if (link.shouldLimitAngle) {
							rotQuat.normalize();
							limitAngleRotation(rotQuat);
						}
						rotQuat.normalize();
						//trace(link.name, rotQuat.toString());
						
						// ボーンへ反映
						link.rotation = link.rotation.multiply(rotQuat);
						link.rotation.normalize();
						
						updateLinksAndTarget(linkIndex);
					}
				}
			}
		}
		
		private function updateLinksAndTarget(top:int):void {
			for (var i:int = top; i >= 0; i--) {
				_links[i].updateTransform();
				//trace(_links[i].name, "updated");
			}
			_target.updateTransform();
		}
		
		
		private function limitAngleRotation(rotation:Quarternion):void {
			var angles:Vector3D = rotation.toEuler();
			if (angles.x < -Math.PI) {
				angles.x = -Math.PI;
			}
			else if ( -0.002 < angles.x) {
				angles.x = -0.002;
			}
			angles.y = 0;
			angles.z = 0;
			rotation.takeIn(angles, Orientation3D.EULER_ANGLES);
		}
	}

}