package net.gamekotan.mmd {
	import flash.events.Event;
	import flash.events.EventDispatcher;
	import flash.events.IOErrorEvent;
	import flash.events.ProgressEvent;
	import flash.geom.Vector3D;
	import flash.net.URLLoader;
	import flash.net.URLLoaderDataFormat;
	import flash.net.URLRequest;
	import flash.utils.ByteArray;
	import flash.utils.Endian;
	
	 /**
	 * Dispatched when the vmd loader completes a file load successfully.
	 * 
	 * @eventType net.gamekotan.mmd.VMDEvent
	 */
	[Event(name="loadSuccess",type="net.gamekotan.mmd.VMDEvent")]
    
	 /**
	 * Dispatched when the vmd loader fails to load a file.
	 * 
	 * @eventType net.gamekotan.mmd.VMDEvent
	 */
	[Event(name="loadError",type="net.gamekotan.mmd.VMDEvent")]
	
	 /**
	 * Dispatched when the vmd loader progresses in the laoding of a file.
	 * 
	 * @eventType net.gamekotan.mmd.VMDEvent
	 */
	[Event(name="loadProgress",type="net.gamekotan.mmd.VMDEvent")]

	/**
	 * ...
	 * @author senzogawa
	 */
	public class VMD extends EventDispatcher {
		
		public var charset:String = "shift_jis";
		private var _loader:URLLoader;
		private var _data:ByteArray;
		
		public var name:String;
		public var motions:Vector.<VMDMotion>;
		public var morphs:Vector.<VMDMorph>;
		public var maxFrame:int;
		
		public function VMD() {
			_loader = new URLLoader();
		}
		
		public function load(url:String):void {
			_loader.dataFormat = URLLoaderDataFormat.BINARY;
			_loader.addEventListener(IOErrorEvent.IO_ERROR, onIOError);
			_loader.addEventListener(ProgressEvent.PROGRESS, onProgress);
			_loader.addEventListener(Event.COMPLETE, onComplete);
			_loader.load(new URLRequest(url));
		}
		
		private function onComplete(event:Event):void {
			_data = _loader.data;
			_data.endian = Endian.LITTLE_ENDIAN;
			dispatchEvent(new VMDEvent(VMDEvent.LOAD_SUCCESS, this));
		}
		
		private function onIOError(event:IOErrorEvent):void {
			dispatchEvent(new VMDEvent(VMDEvent.LOAD_ERROR, this));
		}
		
		private function onProgress(event:ProgressEvent):void {
			dispatchEvent(new VMDEvent(VMDEvent.LOAD_PROGRESS, this));
		}
		
		public function parse():void {
			if (!_data) throw new Error("Illegal State");
			var vmd:ByteArray = _data;
			var header:String = vmd.readMultiByte(30, charset);
			if (header != "Vocaloid Motion Data 0002") throw new Error("Illegal VMD Data");
			name = vmd.readMultiByte(20, charset);
			maxFrame = 0;
			parseMotion(vmd);
			parseMorph(vmd);
			vmd.position = 0;
		}
		
		private function parseMorph(vmd:ByteArray):void {
			var count:int = vmd.readInt();
			//trace("Morphs", count);
			morphs = new Vector.<VMDMorph>(count, true);
			for (var i:int = 0; i < count; i++) {
				morphs[i] = new VMDMorph(
					vmd.readMultiByte(15, charset),
					vmd.readInt(),
					vmd.readFloat()
					);
				maxFrame = Math.max(maxFrame, morphs[i].frame);
			}
		}
		
		private function parseMotion(vmd:ByteArray):void {
			// Motions
			var dummy:ByteArray = new ByteArray();
			var count:int = vmd.readInt();
			//trace("BoneMotions", count);
			motions = new Vector.<VMDMotion>(count, true);
			for (var i:int = 0; i < count; i++) {
				motions[i] = new VMDMotion(
					vmd.readMultiByte(15, charset),
					vmd.readInt(),
					new Vector3D(vmd.readFloat(), vmd.readFloat(), vmd.readFloat()),
					new Vector3D(vmd.readFloat(), vmd.readFloat(), vmd.readFloat(), vmd.readFloat())
					);
				vmd.readBytes(dummy, 0, 64);
				maxFrame = Math.max(maxFrame, motions[i].frame);
			}
		}
		
	}

}