package net.gamekotan.util {
	import flash.geom.Matrix3D;
	import flash.geom.Orientation3D;
	import flash.geom.Vector3D;
	/**
	 * ...
	 * @author senzogawa
	 */
	public class Quarternion {
		
		public static const UNIT:Quarternion = new Quarternion();
		public static const EPSILON:Number = 0.000001;
		
		public var x:Number;
		public var y:Number;
		public var z:Number;
		public var w:Number;
		
		public var matrix:Matrix3D;
		
		public function Quarternion(x:Number = 0, y:Number = 0, z:Number = 0, w:Number = 1) {
			this.x = x;
			this.y = y;
			this.z = z;
			this.w = w;
		}
		
		public static function fromVector3D(vector3d:Vector3D):Quarternion {
			return new Quarternion(vector3d.x, vector3d.y, vector3d.z, vector3d.w);
		}
		
		public function toVector3D():Vector3D {
			return new Vector3D(x, y, z, w);
		}
		
		public function toMatrix3D():Matrix3D {
			var newMatrix:Matrix3D = new Matrix3D();
			var factors:Vector.<Vector3D> = newMatrix.decompose(Orientation3D.QUATERNION);
			var rotation:Vector3D = factors[1];
			rotation.x = x;
			rotation.y = y;
			rotation.z = z;
			rotation.w = w;
			newMatrix.recompose(factors, Orientation3D.QUATERNION);
			return newMatrix;
		}
		
		public function updateMatrix():void {
			var factors:Vector.<Vector3D> = matrix.decompose(Orientation3D.QUATERNION);
			var rotation:Vector3D = factors[1];
			rotation.x = x;
			rotation.y = y;
			rotation.z = z;
			rotation.w = w;
			matrix.recompose(factors, Orientation3D.QUATERNION);
		}
		
		public function toAxisAngle():Vector3D {
			if (w > 1) normalize();
			var angle:Number = 2 * Math.acos(w); // radian
			var sqrt:Number = Math.sqrt(1 - w * w);
			if (sqrt < 0.001) {
				return new Vector3D(x, y, z, angle);
			}
			else {
				return new Vector3D(x / sqrt, y / sqrt, z / sqrt, angle);
			}
		}
		
		public function toEuler():Vector3D {
			var matrix:Matrix3D = toMatrix3D();
			return matrix.decompose(Orientation3D.EULER_ANGLES)[1];
		}
		
		public function takeIn(vector3d:Vector3D, orientation3d:String = null):void {
			if (orientation3d == null) {
				x = vector3d.x;
				y = vector3d.y;
				z = vector3d.z;
				w = vector3d.w;
				return;
			}
			var matrix:Matrix3D = new Matrix3D();
			var factors:Vector.<Vector3D> = matrix.decompose(orientation3d);
			factors[1] = vector3d;
			matrix.recompose(factors, orientation3d);
			var rotation:Vector3D = matrix.decompose(Orientation3D.QUATERNION)[1];
			x = rotation.x;
			y = rotation.y;
			z = rotation.z;
			w = rotation.w;
		}
		
		public function clone():Quarternion {
			return new Quarternion(x, y, z, w);
		}
		
		public function toString():String {
			return "Quarternion(" + x + "," + y + "," + z + "," + w + ")";
		}
		
		public function slerp(another:Quarternion, lerp:Number):Quarternion {
			var qr:Number = x * another.x + y * another.y + z * another.z + w * another.w;
			var t:Number = 1.0 - lerp;
			if (qr < 0) {
				return new Quarternion(
					x * t - another.x * lerp,
					y * t - another.y * lerp,
					z * t - another.z * lerp,
					w * t - another.w * lerp);
			}
			else {
				return new Quarternion(
					x * t + another.x * lerp,
					y * t + another.y * lerp,
					z * t + another.z * lerp,
					w * t + another.w * lerp);
			}
		}
		
		public function normalize():void {
			var sqrt:Number = 1.0 / Math.sqrt(x * x + y * y + z * z + w * w);
			x *= sqrt;
			y *= sqrt;
			z *= sqrt;
			w *= sqrt;
		}
		
		public function multiply(another:Quarternion):Quarternion {
			var qx:Number = another.x;
			var qy:Number = another.y;
			var qz:Number = another.z;
			var qw:Number = another.w;
			return new Quarternion(
				w * qx + x * qw + y * qz - z * qy,
				w * qy - x * qz + y * qw + z * qx,
				w * qz + x * qy - y * qx + z * qw,
				w * qw - x * qx - y * qy - z * qz
			);
		}

	}

}